/* $Id: MakeInharmonicity.java,v 1.12 2017/02/02 08:25:41 mit3 Exp $ */
/* MakeInharmonicity.java v0.2 */

public class MakeInharmonicity {
    private final int KEY_BOARD = 88;
    private final int OCTAVE = 12;
    private final int CURVE = 0; // inha.type
    private final int STRHT = 1;
    private final int LOGG = 2;

    private boolean isCalc;
    private boolean isBend;
    private int ITYPE = CURVE;
    private int WOUND_WIRE = 28;
    private int BEND_KEY = 70;
    private double GRADE_BASS = 0.06;
    private double GRADE_TREBLE = 0.087;
    private double A49I = 0.55;
    private double BEND = 0.087;
    private double[] Inha;

    public MakeInharmonicity() {
	Inha = new double[KEY_BOARD];
    }
    public void setBassGrade(double bg) {
	GRADE_BASS = bg;
    }
    public void setTrebleGrade(double tg) {
	GRADE_TREBLE = tg;
    }
    public void setA49i(double a49i) {
	A49I = a49i;
    }
    public void setWoundWire(int ww) {
	WOUND_WIRE = ww;
    }
    public void setBend(double be) {
	BEND = be;
    }
    public void setBendKey(int bk) {
	BEND_KEY = bk;
    }
    public void setIType(int num) {
	if (num >= 0 && num < 3) 
	    ITYPE = num;
	else
	    System.out.println("*** Inha.type error.:"+num+" ***"); 
    }
    public double[] getInha() {
	double sft = makeTrebleInha();

	switch(ITYPE) {
	case CURVE:
	    makeBassInha(sft);
	    break;
	case STRHT:
	    makeBassStraight(sft);
	    break;
	case LOGG:
	    makeBassLog(sft);
	    break;
	default:
	    makeBassInha(sft);
	    break;
	}
	isCalc = true;

	return Inha;
    }
    public double getInha(int num) {
	if (isCalc)
	    return Inha[num];
	else {
	    System.out.println("*** getInha error.:"+num+" ***");

	    return 0;
	}
    }
    public void resetIsCalc() {
	isCalc = false;
    }
    public boolean isCalc() {
	return isCalc;
    }
    public int getWoundWire() {
	return WOUND_WIRE;
    }
    public void IsBend(boolean sw) {
	isBend = sw;
    }
    public double calcInha(int key) {
	return Math.exp(key*GRADE_TREBLE);
    }
    public double calcInha(int key, double grd) {
	return Math.exp(key*grd);
    }
    public double makeTrebleInha() {
	double df0 = A49I/calcInha(48);
	int n, k = KEY_BOARD - BEND_KEY;
	double gr, sa, df = (BEND - GRADE_TREBLE) / k;
	for (int i = WOUND_WIRE; i < KEY_BOARD; i++) {
	    if (i > BEND_KEY && isBend) {
		n = i - BEND_KEY;
		gr = GRADE_TREBLE + df * n;
		sa = A49I/calcInha(48, gr);
		Inha[i] = calcInha(i, gr)*sa;
	    } else
		Inha[i] = calcInha(i)*df0;
        }

	return Inha[WOUND_WIRE]; // Ǿ᤹ͤ
    }
    public void makeBassStraight(double sft) {
	double[] inha = new double[WOUND_WIRE+1];

	for (int i = 0; i < WOUND_WIRE+1; i++) {
	    inha[i] = 1.0/calcInha(i, GRADE_BASS)*0.8;
        }

	addInha(inha, sft);
    }
    public void makeBassLog(double sft) {
	double[] inha = new double[WOUND_WIRE+1];

	for (int i = 0; i < WOUND_WIRE+1; i++) {
	    inha[i] = Math.pow(i+1, -GRADE_BASS*6.5);
	}

	addInha(inha, sft);
    }
    public void addInha(double[] inha, double sft) {
	if (WOUND_WIRE < 1)
	    return;

	double df = inha[WOUND_WIRE]/sft;

	for (int i = 0; i < WOUND_WIRE; i++) {
	    Inha[i] = inha[i]/df;
	}
    }
    public void makeBassInha(double sft) { // curve
	double[] inha = new double[WOUND_WIRE+1];

	for (int i = 0; i < WOUND_WIRE+1; i++) {
	    inha[i] = cosh(GRADE_BASS*(i-WOUND_WIRE));
	}

	addInha(inha, sft);
    }
    public double cosh(double x) {
        double me = Math.pow(Math.E, x);

        return (me+1/me)/2;
    }
    //
    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append("Grade bass:"+(float)GRADE_BASS);
	sb.append(" Grade treble:"+(float)GRADE_TREBLE);
        sb.append(" Pitch:"+(float)A49I);
        sb.append(" Wound Wire:"+WOUND_WIRE);

	return sb.toString();
    }
}
