/* $Id: TuningCenter.java,v 1.6 2016/08/27 22:54:50 mit3 Exp $ */
/* TuningCenter v0.1.2 */
import java.text.DecimalFormat;

public class TuningCenter {
    private final int KEY_BOARD = 88;
    private final int OCTAVE = 12;
    private final int OCTAVE2 = 24;
    private final int P4 = 5;
    private final int P5 = 7;
    private final int NONZERO = 0; // TUNE
    private final int EXT_OCT = 1;
    private final int EXT2OCT = 2;
    private final int CURVE = 3;
    private final int ENVELOPE = 4;
    private final int TRIADS = 5;
    private final int CLEAR = 6;
    private final int CENT = 1200;
    private final double REVOCT = 1.0/OCTAVE;
    private final double SAQ = Math.pow(22/8.0, 1.0/(KEY_BOARD-48));
    private final int[] KEYS = {36,41,34,39,32,37,42,35,40,33,38,39};
    private final int[] LAYS = {P4,-P5,P4,-P5,P4,P4,-P5,P4,-P5,P4,P4,P4};

    private boolean isCalc;
    private boolean isLaying;
    private boolean AUTO_BEAT;
    private boolean AUTO_CENT;
    private boolean DEBUG;
    private int FROM_KEY = 33-1;  // f
    private int TO_KEY = 45-1;    // f
    private int WOUND_WIRE = 28;  // 28-1
    private int TUNE = NONZERO;
    private int OCT_TYPE = OCTAVE;
    private double BEAT = 1.0;
    private double BEAT4th = 1.098;//1.13; 
    private double BEAT5th = 0.394;//0.36;
    private double[] Cent;
    private double[] Inha;

    private TuningFormula2 TF2;
    private TuningTriads TT;
    private TuningEnvelope TE;
    private TuningNonZero TN;
    private DecimalFormat df1;
    private Interval I;
    private DebugData[] DD;
    private MinMax MM;

    public TuningCenter(boolean debug) {
	DEBUG = debug;
	df1 = new DecimalFormat("#0.0##");
	I = new Interval();
	TF2 = new TuningFormula2();
	TT = new TuningTriads();
	TE = new TuningEnvelope();
	TN = new TuningNonZero(DEBUG);

	resetCent();
    }
    //
    public void setInha(double[] inha) {
	Inha = inha;
    }
    public void Tuning() {
	if (DEBUG)
	    DD = new DebugData[KEY_BOARD];

	resetCent();
	tuneLaying();  // disp_laying();
	
	switch (TUNE) {
	case TRIADS:
	    resetAutoBeat();
	    TT.setInha(Inha);
	    TT.setCent(Cent);
	    TT.Tuning();
	    Cent = TT.getCent();
	    break;
	case ENVELOPE:
	    resetAutoBeat();
	    TE.setInha(Inha);
	    TE.setCent(Cent);
	    TE.Tuning();
	    Cent = TE.getCent();
	    break;
	case NONZERO:
	    resetAutoBeat();
	    TN.setInha(Inha);
	    TN.setCent(Cent);
	    TN.setBeat(BEAT);
	    TN.Tuning();
	    Cent = TN.getCent();
	    break;
	case EXT_OCT:
	    tuneExtension();
	    break;
	case EXT2OCT:
	    tune2octave();
	    break;
	case CURVE:
	    resetAutoBeat();
	    TF2.getInharmonicity(Inha);
	    Cent = TF2.getCurve(BEAT);
	    break;
	case CLEAR:
	    resetAutoBeat();
	    resetCent();
	    break;
	default:
	    break;
	}
	isCalc = true;
    }
    public double[] getCent() {
	return Cent;
    }
    public double getCent(int num) {
	return Cent[num];
    }
    public double getICent(int num) {
	return Inha[num]+Cent[num];
    }
    public double getICent(int key, int par) {
	return Inha[key]*par*par+Cent[key];
    }
    public void resetCent() {
	Cent = new double[KEY_BOARD];

	for (int i = 0; i < KEY_BOARD; i++)
	    Cent[i] = 0;
    }
    public int[] getKeys() {
	return KEYS;
    }
    public int[] getLays() {
	return LAYS;
    }
    public DebugData getDD(int key) {
	return TN.getDD(key);
    }
    public void setTune(int num) {
	if (num < NONZERO || num > CLEAR) {
	    System.out.println("*** TUNE error:"+num+" ***");
	    TUNE = -1;
	} else
	    TUNE = num;
    }
    public int getTune() {
	return TUNE;
    }
    public void setBeat(double beat) {
	BEAT = beat;

	this.resetIsCalc();
    }
    public double getBeat() {
	return BEAT;
    }
    public void setBeat4th(double beat) {
	BEAT4th = beat;

	if (AUTO_CENT)
	    calcDiff(P5);

	this.resetIsCalc();
    }
    public double getBeat4th() {
	return BEAT4th;
    }
    public void setBeat5th(double beat) {
	BEAT5th = beat;

	if (AUTO_CENT)
	    calcDiff(P4);

	this.resetIsCalc();
    }
    public double getBeat5th() {
	return BEAT5th;
    }
    public void resetIsCalc() {
	isCalc = false;
    }
    public boolean isCalc() {
	return isCalc;
    }
    public void setLaying(boolean sw) {
	isLaying = sw;
    }
    public boolean isLaying() {
	return isLaying;
    }
    public void setAutoCent(boolean sw) {
	AUTO_CENT = sw;
    }
    public void setAutoBeat(boolean sw) {
	AUTO_BEAT = sw;
    }
    public void resetAutoBeat() {
	AUTO_BEAT = false;
    }
    public boolean isAutoBeat() {
	return AUTO_BEAT;
    }
    public void setOctaveType(int num) {
	OCT_TYPE = num;
	if (OCT_TYPE > 14)
	    OCT_TYPE = 14;
	else if (OCT_TYPE < 12)
	    OCT_TYPE = 12;
    }
    public int isOctaveType() {
	return OCT_TYPE;
    }
    public void setWoundWire(int ww) {
	WOUND_WIRE = ww;

	this.resetIsCalc();
    }
    //
    public void ff_beat() { // for debug
	int[] ratio = I.getIratio(OCTAVE);
	double beat = get_ibeats(TO_KEY, FROM_KEY, ratio);

	System.out.println("ff_beat:"+(float)beat);
    }
    public void beat_setO() {
	int[] ratio = I.getIdata(OCT_TYPE);
	BEAT = get_ibeats(TO_KEY, FROM_KEY, ratio);
    }
    public void beat_set2O() {
	int[] ratio = I.getIratio(OCTAVE2);
	BEAT = get_ibeats(TO_KEY+OCTAVE, FROM_KEY, ratio);
    }
    //
    public void tuneLaying() {
	int key0, key1;
	double be;

	for (int i = 0; i < OCTAVE; i++) {
            key0 = KEYS[i];
            key1 = key0+LAYS[i];
            be = (Math.abs(LAYS[i]) == P4) ? BEAT4th : BEAT5th;
            tuneBeat(key0, key1, be);
        }
    }
    public double diffCent() {
	double sm = 0;
	for (int i = FROM_KEY; i < TO_KEY; i++)
	    sm += Math.abs(Cent[i]-Cent[i+1]);

	return sm;
    }
    public void calcDiff(int sw) { // auto_cent for 4th&5th
	double sa = 0.01;
	double ce = diff_th(sw, 0);
	double cp = diff_th(sw, sa);

	if (cp < ce) {
	    while (cp < ce) {
		ce = cp;
		cp = diff_th(sw, sa);
	    }
	} else {
	    pm_beat(sw, -sa);
	    while (ce < cp) {
		cp = ce;
		ce = diff_th(sw, -sa);
	    }
	}
    }
    public double diff_th(int sw, double dc) {
	pm_beat(sw, dc);
	tuneLaying();

	return diffCent();
    }
    public void pm_beat(int sw, double ce) {
	if (sw == P4)
	    BEAT4th += ce;
	else
	    BEAT5th += ce;
    }
    //
    public void tuneBeat(int key0, int key1, double beat) {
	double be = calcBeat(key0, key1);
	double sa = Math.abs(beat-be);

	while (sa > 0.005) {
	    sa = Math.abs(beat-be);
	    be = calc_beat(key0, key1, beat);
	}
    }
    public double calc_beat(int k0, int k1, double beat) {
	double dc = 0.01;
	double b0 = calcBeat(k0, k1);
	double sa = Math.abs(b0-beat);
	if (sa > 0.3)
	    dc = 0.5; //
	else if (sa > 0.1)
	    dc = 0.05;
	else
	    dc = 0.01;

	Cent[k1] = (b0 < beat) ? Cent[k1]+dc : Cent[k1]-dc;
	return b0;
    }
    public double calcBeat(int key0, int key1) {
	int mul = 5;
	double[] f0 = calcPartial(mul, key0);
	double[] f1 = calcPartial(mul, key1);
	double[] be = new double[mul];

	double min = Double.MAX_VALUE;
	for (int i = 0; i < mul; i++) {
	    for (int j = 0; j < mul; j++) {
		be[j] = f1[j]-f0[i];
		if (Math.abs(min) > Math.abs(be[j]))
		    min = be[j];
	    }
	}

	return min;
    }
    public double[] calcPartial(int mul, int key) {
	double[] f = new double[mul];

	for (int i = 0; i < mul; i++)
	    f[i] = calcFreq(i+1, key);

	return f;
    }
    public double calcFreq(int par, int key) {
	return ctof(getFreq(key)*par, getICent(key, par));
    }
    public void disp_laying() { // for debug
	int key0, key1;
	double be4, be5, ic;

	System.out.println("Key#: 4th : 5th : iCent : Inha. : Freq.");
	System.out.println("---------------------------------------");
	for (int i = FROM_KEY; i < TO_KEY; i++) {
	    key0 = i+P4;
	    key1 = i+P5;
	    be4 = calcBeat(i, key0);
	    be5 = calcBeat(i, key1);
	    ic = getICent(i);

	    System.out.println((i+1)+": "+dform(be4)+" "+dform(be5)+
			       " "+dform(ic)+" "+dform(Inha[i])+
			       " "+dform(calcFreq(1, i)));
	}
    }
    public double ctof(double f, double c) {
        return f*Math.pow(2, c/CENT);
    }
    public double getFreq(int key) {
	return 440*Math.pow(2, (key-48)*REVOCT); // 49-1
    }
    //
    public double calcSindo(double freq, int pr, int key) {
	double fg = getSboard(freq);
	double sp = getSpoint(key);
	double ga = spoint(pr, sp);

	return ga * fg;
    }
    public double getSpoint(int key) { // 0~481/8
	return (key < 49) ? 8 : 8 * Math.pow(SAQ, key-48); // 0~87
    }
    public double spoint(int mm, double sp) {
	double mp = mm*Math.PI;
	return Math.abs(8.0*Math.sin(mp/sp)/mp);
    }
    public double getSboard(double freq) {
	freq -= 20;
	double gain = freq/170;

	if (gain < 0)
	    gain = 0.0;
	else if (gain > 1.0)
	    gain = 1.0;

	return gain;
    }
    //
    public double get_mbeats(int keyh, int keyl, int[] ratio, int mul) {
	return get_ibeats(keyh, keyl, calc_multi(ratio, mul));
    }
    public double getMbeats(int key, int inte, int mul) { // jts
	int[] ratio = I.getIdata(inte); // -б
	int keyh = key+ratio[2];

	return get_mbeats(keyh, key, ratio, mul);
    }
    public int[] calc_multi(int[] ratio, int mul) {
	int[] rat = new int[3];

	if (mul < 1) {
	    System.out.println("*** Multi input error. ***");
	    mul = 1;
	}
	rat[0] = ratio[0]*mul; // low
	rat[1] = ratio[1]*mul; // hi
	rat[2] = ratio[2];

	return rat;
    }
    public double get_ibeats(int keyh, int keyl, int[] ratio) {
	double fh = calcFreq(ratio[1], keyh);
	double fl = calcFreq(ratio[0], keyl);

	return (fh-fl);
    }
    //
    public void tune2octave() { // 2octave
	int[] ratio = I.getIratio(OCTAVE);
	int [] ratio2 = I.getIratio(OCTAVE*2);

	if (AUTO_BEAT)
	    beat_set2O();

	for (int i = TO_KEY+1; i < KEY_BOARD; i++) {
	    Cent[i] = Cent[i-1];
	    if (i < TO_KEY+OCTAVE)
	    	tune_ext_treble(i, ratio);
	    else
		tune_ext_treble(i, ratio2);
	}
	for (int i = FROM_KEY-1; i >= 0; i--) {
	    Cent[i] = Cent[i+1];
	    tune_ext_bass(i, ratio2);
	}
    }
    //
    public void tuneExtension() { // octave
	if (AUTO_BEAT)
	    beat_setO();

	int[] ratio = I.getIdata(OCT_TYPE); // 2:1,4:2,6:3
	for (int i = TO_KEY+1; i < KEY_BOARD; i++) {
	    tune_ext_treble(i, ratio);
	}

	for (int i = FROM_KEY-1; i >= 0; i--) {
	    tune_ext_bass(i, ratio);
	}
    }
    public void tune_ext_treble(int key, int[] ratio) {
	int keyl = key-ratio[2];
	double dc = 0.1;
	double be0 = 0;
	double be = ext_beat(key, keyl, ratio);

	do {
	    Cent[key] += dc;
	    be0 = be;
	    be = ext_beat(key, keyl, ratio);
	} while (be < be0);

	do {
	    Cent[key] -= dc;
	    be0 = be;
	    be = ext_beat(key, keyl, ratio);
	} while (be < be0);
    }
    public double ext_beat(int keyh, int keyl, int[]ratio) {
	double be = get_ibeats(keyh, keyl, ratio);

	return Math.abs(be-BEAT);
    }
    public void tune_ext_bass(int key, int[] ratio) {
	int keyh = key+ratio[2];
	double dc = -0.1;
	double be0;
	double be = ext_beat(keyh, key, ratio);

	do {
	    Cent[key] -= dc;
	    be0 = be;
	    be = ext_beat(keyh, key, ratio);
	} while (be < be0);

	do {
	    Cent[key] += dc;
	    be0 = be;
	    be = ext_beat(keyh, key, ratio);
	} while (be < be0);
    }
    //
    public String toString() {
        StringBuffer sb = new StringBuffer();

	sb.append("FromKey:"+FROM_KEY);
        sb.append(" ToKey:"+TO_KEY);
        sb.append(" WoundWire:"+WOUND_WIRE);

	return sb.toString();
    }
    public String dform(double n) {
        return df1.format((float)n);
    }
}
