/* $Id: TuningEnvelope.java,v 1.4 2016/11/07 23:30:06 mit3 Exp $ */
/* TuningEnvelope v0.1.1 */

public class TuningEnvelope {
    private final int KB = 88;
    private final int P4 = 5;
    private final int OCTAVE = 12;
    private final int OCTAVE2 = 24;
    private final int CENT = 1200;
    private final double PITCH = 440;
    private final double REVOCT = 1.0/OCTAVE;
    private final double SAQ = Math.pow(22/8.0, 1.0/(KB-48));

    private int[] OCT;
    private int[] OCT2;
    private int[] P4TH;
    private double[] Cent;
    private double[] Inha;
    private Interval I;

    public TuningEnvelope() {
	I = new Interval();
	OCT = I.getIdata(OCTAVE); // [2, 1, 12]
	OCT2 = I.getIratio(OCTAVE2);
	P4TH = I.getIratio(P4);
    }
    public void setInha(double[] inha) {
        Inha = inha;
    }
    public void setCent(double[] cent) {
        Cent = cent;
    }
    public double[] getCent() {
        return Cent;
    }
    public void Tuning() {
	tuneBassE();
	tuneTrebleE();
    }
    public void tuneBassE() {
        for (int i = 31; i >= 0; i--) {
            tune_bassE(i);
        }
    }
    public void tune_bassE(int key) {
        int keyh = key + OCT[2];
	double be, minc = 0, minb = Double.MAX_VALUE;
	double dc = -0.1;
	double max = 60;

	while (true) {
	    Cent[key] -= dc;
	    be = Math.abs( partial_bassE(keyh, key) );
	    if (minb > be) {
                minb = be;
                minc = Cent[key];
            }
	    if (Cent[key] > max)
		break;
	}
	while (true) {
	    Cent[key] += dc;
	    be = Math.abs( partial_bassE(keyh, key) );
	    if (minb > be) {
                minb = be;
                minc = Cent[key];
            }
	    if (Cent[key] < -max)
		break;
	}

	Cent[key] = minc;
    }
    public double partial_bassE(int keyh, int keyl) {
	double sm = 0;
        int cnt = 1;
        int pmax = 12;//10;

        for (int i = 1; i <= pmax; i++) {
            if (i % OCT[0] == 0) {
                sm += calc_sum(OCT, cnt, keyl, keyh);
                cnt += 1;
            }
        }
        return sm;
    }
    //
    public void tuneTrebleE() {
        for (int i = 45; i < KB; i++) {
            tune_trebleE(i);
        }
    }
    public void tune_trebleE(int key) {
        double be, minc = 0, minb = Double.MAX_VALUE;
        double dc = 0.1;
	double max = 60;

	while (true) {
	    Cent[key] -= dc;
            be = Math.abs( get3KeyBeatE(key) );

	    if (minb > be) {
                minb = be;
                minc = Cent[key];
            }
	    if (Cent[key] < -max)
		break;
	}
	while (true) {
	    Cent[key] += dc;
            be = Math.abs( get3KeyBeatE(key) );

	    if (minb > be) {
                minb = be;
                minc = Cent[key];
            }
	    if (Cent[key] > max)
		break;
	}

	Cent[key] = minc;
    }
    public double get3KeyBeatE(int key) {
        if (key > 44+OCTAVE) {
            double be0 = calc_sum(OCT, 1, key - OCTAVE, key);
            double be1 = calc_sum(OCT2, 1, key - OCTAVE2, key);
            return Math.abs(be0 + be1); // abs
        } else { // ǽΥ֤ 4th & oct 
            double b4 = calc_sum(P4TH, 1, key - P4, key);
            double b0 = calc_sum(OCT, 1, key - OCTAVE, key);
            return Math.abs(b4 + b0 - 1.1);
        }
    }
    //
    public double calc_sum(int[] rat, int cnt, int keyl, int keyh) {
        double sm = 0;
        int ral = rat[0]*cnt;
        int rah = rat[1]*cnt;

        double frl = calcFreq(ral, keyl);
        double frh = calcFreq(rah, keyh);
        double be = frh-frl;

        double gal = calcSindo(frl, ral, keyl);
        double gah = calcSindo(frh, rah, keyh);
        double gain = gal+gah;

        sm = be*gain;

        return sm;
    }
    public double calcFreq(int par, int key) {
        return ctof(getFreq(key)*par, getICent(key, par));
    }
    public double ctof(double f, double c) {
        return f*Math.pow(2, c/CENT);
    }
    public double getFreq(int key) {
        return PITCH*Math.pow(2, (key-48)*REVOCT); // 49-1
    }
    public double getICent(int key, int par) {
        return Inha[key]*par*par+Cent[key];
    }
    public double calcSindo(double freq, int pr, int key) {
        double fg = getSboard(freq);
        double sp = getSpoint(key);
        double ga = spoint(pr, sp);

        return ga * fg;
    }
    public double getSpoint(int key) { // 0~481/8
        return (key < 49) ? 8 : 8 * Math.pow(SAQ, key-48); // 0~87
    }
    public double spoint(int mm, double sp) {
        double mp = mm*Math.PI;
        return Math.abs(8.0*Math.sin(mp/sp)/mp);
    }
    public double getSboard(double freq) {
        freq -= 20;
        double gain = freq/170;

        if (gain < 0)
            gain = 0.0;
        else if (gain > 1.0)
            gain = 1.0;

        return gain;
    }
}
