/* $Id: TuningNonZero.java,v 1.74 2016/11/07 23:31:06 mit3 Exp $ */
/* TuningNonZero v0.5 */

public class TuningNonZero {
    private final int KEY_BOARD = 88;
    private final int OCTAVE = 12;
    private final int OCTAVE2 = 24;
    private final int P4 = 5;
    private final int CENT = 1200;
    private final double REVOCT = 1.0/OCTAVE;
 
    private boolean DEBUG;
    private int FROM_KEY = 33-1;  // f
    private int TO_KEY = 45-1;    // f
    private int WOUND_WIRE = 28;  // 28-1
    private double BEAT = 1.0;
    private double[] Cent;
    private double[] Inha;

    private Interval I;
    private DebugData[] DD;
    private MinMax MM;

    public TuningNonZero(boolean debug) {
	I = new Interval();

	DEBUG = debug;
    }
    //
    public void setInha(double[] inha) {
	Inha = inha;
    }
    public void setCent(double[] cent) {
	Cent = cent;
    }
    public void Tuning() {
	if (DEBUG)
	    DD = new DebugData[KEY_BOARD];

	tuneBass();
	tuneTreble();
    }
    public double[] getCent() {
	return Cent;
    }
    //
    public double getICent(int num) {
	return Inha[num]+Cent[num];
    }
    public double getICent(int key, int par) {
	return Inha[key]*par*par+Cent[key];
    }
    public void resetCent() {
	Cent = new double[KEY_BOARD];
    }
    //
    public DebugData getDD(int key) {
	return DD[key];
    }
    //
    public void setBeat(double beat) {
        BEAT = beat;
    }
    public void setWoundWire(int ww) {
	WOUND_WIRE = ww;
    }
   
    public double calcFreq(int par, int key) {
	return ctof(getFreq(key)*par, getICent(key, par));
    }

    public double ctof(double f, double c) {
        return f*Math.pow(2, c/CENT);
    }
    public double getFreq(int key) {
	return 440*Math.pow(2, (key-48)*REVOCT); // 49-1
    }
    //
    public void tuneBass() {
	int[] ratio = I.getIratio(OCTAVE);
	for (int i = FROM_KEY-1; i >= 0; i--) {
	    tune_bass(i, ratio);
	}
    }
    public void tune_bass(int key, int[] ratio) {
	int keyh = key + ratio[2];
	double be = 0;
	double dc = -0.1;
	double minc = 0;
	double minb = 10000;
	int max = 60;
	int len = (int)(60/-dc);
	int num = 0;
	MM = new MinMax(DEBUG, len);

	while(true) {
	    be = Math.abs( partial_bass(keyh, key, ratio) );
	    MM.minmax(be, Cent[key], num++);
	    if (be < minb) {
		minb = be;
		minc = Cent[key];
	    }
	    Cent[key] -= dc;
	    if (Cent[key] > max)
		break;
	}
	while(true) {
	    be = Math.abs( partial_bass(keyh, key, ratio) );
	    MM.minmax(be, Cent[key], num++);
	    if (be < minb) {
		minb = be;
		minc = Cent[key];
	    }
	    Cent[key] += dc;
	    if (Cent[key] < -max)
		break;
	}

	Cent[key] = minc;

	if (DEBUG) {
	    DD[key] = new DebugData(key, Cent[key], len, MM);//ch, MM);
	}
    }
    public double partial_bass(int keyh, int keyl, int[] ratio) {
	int ww = WOUND_WIRE-1; // -1Ǵ̵

	if (ww >= 0 && keyl < WOUND_WIRE) {
	    int len = 5;
	    double df = 0;
	    double[] be = new double[len];

	    for (int i = 0; i < 5; i++)
		be[i] = get_mbeats(keyh, keyl, ratio, i+1);
	    for (int j = 0; j < len-1; j++)
		df += Math.abs(Math.abs(be[j+1])-Math.abs(be[j]));

	    return df;
	} else {
	    int[] rat4th = I.getIratio(P4);
	    double b4 = get_ibeats(keyl+rat4th[2], keyl, rat4th);

	    return Math.abs(b4-BEAT);
	}
    }
    public double get_mbeats(int keyh, int keyl, int[] ratio, int mul) {
	return get_ibeats(keyh, keyl, calc_multi(ratio, mul));
    }
    public double getMbeats(int key, int inte, int mul) {
	int[] ratio = I.getIdata(inte); // -б
	int keyh = key+ratio[2];

	return get_mbeats(keyh, key, ratio, mul);
    }
    public int[] calc_multi(int[] ratio, int mul) {
	int[] rat = new int[3];

	if (mul < 1) {
	    System.out.println("*** Multi input error. ***");
	    mul = 1;
	}
	rat[0] = ratio[0]*mul; // low
	rat[1] = ratio[1]*mul; // hi
	rat[2] = ratio[2];

	return rat;
    }
    public double get_ibeats(int keyh, int keyl, int[] ratio) {
	double fh = calcFreq(ratio[1], keyh);
	double fl = calcFreq(ratio[0], keyl);

	return (fh-fl);
    }
    public void tuneTreble() {
	int[] ratio = I.getIratio(OCTAVE);
	for (int i = TO_KEY+1; i < KEY_BOARD; i++) {
	    tune_treble(i, ratio);
	}
    }
    public void tune_treble(int key, int[] ratio) {
	int keyl = key-ratio[2];
	double dc = 0.1; //0.01;
	double minc = 0;
	double minb = 10000;
	double be;
	int max = 60;

	while(true) {
	    be = Math.abs( partial_high(key, keyl, ratio) );
	    if (be < minb) {
		minb = be;
		minc = Cent[key];
	    }
	    if (Cent[key] < -max)
		break;
	    Cent[key] -= dc;
	}
	while(true) {
	    be = Math.abs( partial_high(key, keyl, ratio) );
	    if (be < minb) {
		minb = be;
		minc = Cent[key];
	    }
	    if (Cent[key] > max)
		break;
	    Cent[key] += dc;
	}

	Cent[key] = minc;
    }
    public double ext_beat(int keyh, int keyl, int[]ratio) {
        double be = get_ibeats(keyh, keyl, ratio);

        return Math.abs(be-BEAT);
    }
    public double partial_high(int keyh, int keyl, int[]ratio) {
	if (keyh > TO_KEY+OCTAVE)
	    return get3KeyBeat(keyh);
	else {
	    int[] rat4th = I.getIratio(P4);
	    double b4 = get_ibeats(keyh, keyh-rat4th[2], rat4th);
	    return Math.abs(b4-BEAT);
	}
    }
    public double get3KeyBeat(int key) {
	double f0 = calcFreq(1, key);
	double be0 = f0-calcFreq(2, key-OCTAVE);   // octave
	double be1 = f0-calcFreq(4, key-OCTAVE*2); // 2octave
  
	return Math.abs(be0+be1);
    }
}
